# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

import logging
from abc import ABC, abstractmethod
from datetime import datetime
from functools import lru_cache
from typing import Any, Callable, Generic, Iterable, List, Mapping, MutableMapping, Optional, Tuple, Type, TypeVar, Union

from source_faker.canon.canonical_model import CanonicalModel

from airbyte_cdk import (
    AirbyteMessage,
    AirbyteStream,
    ConfiguredAirbyteStream,
    ConnectorStateManager,
    Cursor,
    InternalConfig,
    Record,
    Stream,
    StreamFacade,
    SyncMode,
)
from airbyte_cdk.models import (
    AirbyteLogMessage,
    AirbyteMessage,
    AirbyteStream,
    ConfiguredAirbyteStream,
    Level,
    SyncMode,
)
from airbyte_cdk.models import (
    Type as MessageType,
)
from airbyte_cdk.sources import AbstractSource, Source
from airbyte_cdk.sources.streams.concurrent.abstract_stream import AbstractStream
from airbyte_cdk.sources.streams.concurrent.abstract_stream_facade import AbstractStreamFacade
from airbyte_cdk.sources.streams.concurrent.availability_strategy import StreamAvailability
from airbyte_cdk.sources.streams.concurrent.default_stream import DefaultStream
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from airbyte_cdk.sources.streams.core import StreamData
from airbyte_cdk.sources.types import Record
from airbyte_cdk.sources.utils.slice_logger import SliceLogger


T = TypeVar("T", bound=CanonicalModel)


# This is also a callable...
class TransformFunction(Generic[T], ABC):
    @abstractmethod
    def __call__(self, record: Record) -> Record:
        pass


class CanonicalPartition(Partition):
    def __init__(self, canonical_model_type: Type[T], transform_function: TransformFunction[T], partition: Partition):
        self._canonical_model_type = canonical_model_type
        self._transform_function = transform_function
        self._partition = partition

    def read(self) -> Iterable[Record]:
        for record in self._partition.read():
            canonical_record = self._transform_function(record)
            self._canonical_model_type.validate(canonical_record.data)
            self._datetime_to_string(canonical_record.data)
            yield canonical_record

    def to_slice(self) -> Optional[Mapping[str, Any]]:
        return self._partition.to_slice()

    def stream_name(self) -> str:
        return self._canonical_model_type.stream_name()

    def cursor_field(self) -> Optional[str]:
        return "updated_at"

    def __hash__(self) -> int:
        return hash(self._partition)

    def _datetime_to_string(self, data: Any) -> Any:
        if isinstance(data, dict):
            for key, value in data.items():
                data[key] = self._datetime_to_string(value)
        elif isinstance(data, list):
            for item in data:
                item = self._datetime_to_string(item)
        elif isinstance(data, datetime):
            return data.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        return data


class CanonicalStream(AbstractStream):
    def __init__(
        self,
        stream: AbstractStream,
        canonical_model_type: Type[T],
        transform_function: TransformFunction[T],
        cursor_override: Optional[Cursor] = None,
        namespace: Optional[str] = None,
        logger: logging.Logger = logging.getLogger("airbyte"),
    ):
        self._stream = stream
        self._canonical_model_type = canonical_model_type
        self._transform_function = transform_function
        self._namespace = namespace
        self._logger = logger
        self._cursor = cursor_override

    def generate_partitions(self) -> Iterable[Partition]:
        for partition in self._stream.generate_partitions():
            yield CanonicalPartition(self._canonical_model_type, self._transform_function, partition)

    @property
    def cursor(self) -> Cursor:
        if self._cursor is not None:
            return self._cursor
        else:
            return self._stream.cursor

    @property
    def name(self) -> str:
        return self._canonical_model_type.stream_name()

    @property
    def cursor_field(self) -> Optional[str]:
        return "updated_at"

    def check_availability(self) -> StreamAvailability:
        return self._stream.check_availability()

    def get_json_schema(self) -> Mapping[str, Any]:
        return self._canonical_model_type.model_json_schema()

    def as_airbyte_stream(self) -> AirbyteStream:
        stream = AirbyteStream(
            name=self.name,
            json_schema=dict(self._canonical_model_type.model_json_schema()),
            supported_sync_modes=[SyncMode.full_refresh],
            is_resumable=False,
        )

        if self._namespace:
            stream.namespace = self._namespace

        if self.cursor_field:
            stream.source_defined_cursor = True
            stream.is_resumable = True
            stream.supported_sync_modes.append(SyncMode.incremental)
            stream.default_cursor_field = [self.cursor_field]

        keys = ["id"]
        if keys and len(keys) > 0:
            stream.source_defined_primary_key = [[key] for key in keys]

        return stream

    def log_stream_sync_configuration(self) -> None:
        self._logger.debug(
            f"Syncing stream instance: {self.name}",
            extra={
                "primary_key": ["id"],
                "cursor_field": self.cursor_field,
            },
        )


def create_canonical_stream(
    stream: AbstractStream,
    canonical_model_type: Type[T],
    transform_function: TransformFunction[T],
    cursor_override: Optional[Cursor],
    namespace: Optional[str] = None,
) -> CanonicalStream:
    return CanonicalStream(stream, canonical_model_type, transform_function, cursor_override, namespace)


class CanonicalStreamFacade(AbstractStreamFacade[CanonicalStream], Stream):
    def __init__(
        self,
        stream: CanonicalStream,
        cursor: Cursor,
        slice_logger: SliceLogger,
        logger: logging.Logger,
    ):
        self._stream = stream
        self._cursor = cursor
        self._slice_logger = slice_logger
        self._logger = logger

    @property
    def state(self) -> MutableMapping[str, Any]:
        raise NotImplementedError("This should not be called as part of the Concurrent CDK code. Please report the problem to Airbyte")

    @state.setter
    def state(self, value: Mapping[str, Any]) -> None:
        pass

    def read(
        self,
        configured_stream: ConfiguredAirbyteStream,
        logger: logging.Logger,
        slice_logger: SliceLogger,
        stream_state: MutableMapping[str, Any],
        state_manager: ConnectorStateManager,
        internal_config: InternalConfig,
    ) -> Iterable[StreamData]:
        yield from self._read_records()

    def read_records(
        self,
        sync_mode: SyncMode,
        cursor_field: Optional[List[str]] = None,
        stream_slice: Optional[Mapping[str, Any]] = None,
        stream_state: Optional[Mapping[str, Any]] = None,
    ) -> Iterable[StreamData]:
        try:
            yield from self._read_records()
        except Exception as exc:
            if hasattr(self._cursor, "state"):
                state = str(self._cursor.state)
            else:
                # This shouldn't happen if the ConcurrentCursor was used
                state = "unknown; no state attribute was available on the cursor"
            yield AirbyteMessage(
                type=MessageType.LOG,
                log=AirbyteLogMessage(level=Level.ERROR, message=f"Cursor State at time of exception: {state}"),
            )
            raise exc

    def _read_records(self) -> Iterable[StreamData]:
        for partition in self._stream.generate_partitions():
            if self._slice_logger.should_log_slice_message(self._logger):
                yield self._slice_logger.create_slice_log_message(partition.to_slice())
            for record in partition.read():
                yield record.data

    @property
    def name(self) -> str:
        return self._stream.name

    @property
    def primary_key(self) -> Optional[Union[str, List[str], List[List[str]]]]:
        # This method is not expected to be called directly. It is only implemented for backward compatibility with the old interface
        return self.as_airbyte_stream().source_defined_primary_key  # type: ignore # source_defined_primary_key is known to be an Optional[List[List[str]]]

    @property
    def cursor_field(self) -> Union[str, List[str]]:
        if self._stream.cursor_field is None:
            return []
        else:
            return self._stream.cursor_field

    @property
    def cursor(self) -> Optional[Cursor]:  # type: ignore[override] # StreamFaced expects to use only airbyte_cdk.sources.streams.concurrent.cursor.Cursor
        return self._cursor

    @lru_cache(maxsize=None)
    def get_json_schema(self) -> Mapping[str, Any]:
        return {}  # FIXME

    @property
    def supports_incremental(self) -> bool:
        return True  # FIXME

    def check_availability(self, logger: logging.Logger, source: Optional["Source"] = None) -> Tuple[bool, Optional[str]]:
        """
        Verifies the stream is available. Delegates to the underlying AbstractStream and ignores the parameters
        :param logger: (ignored)
        :param source:  (ignored)
        :return:
        """
        availability = self._stream.check_availability()
        return availability.is_available(), availability.message()

    def as_airbyte_stream(self) -> AirbyteStream:
        return self._stream.as_airbyte_stream()

    def log_stream_sync_configuration(self) -> None:
        self._stream.log_stream_sync_configuration()

    def get_underlying_stream(self) -> CanonicalStream:
        return self._stream


def create_canonical_stream_facade(
    stream: AbstractStream,
    canonical_model_type: Type[T],
    transform_function: TransformFunction[T],
    cursor_override: Optional[Cursor],
    logger: logging.Logger,
    slice_logger: SliceLogger,
    namespace: Optional[str] = None,
) -> CanonicalStreamFacade:
    canonical_stream = create_canonical_stream(stream, canonical_model_type, transform_function, cursor_override, namespace)
    return CanonicalStreamFacade(canonical_stream, canonical_stream.cursor, slice_logger, logger)



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from datetime import datetime
from typing import Literal

from pydantic import BaseModel, Field
from source_faker.canon.canonical_model import CanonicalModel, TimestampedModel


AddressType = Literal["BILLING", "SHIPPING", "OTHER"]
EmailType = Literal["WORK", "PERSONAL", "OTHER"]
PhoneType = Literal["WORK", "PERSONAL", "OTHER"]


class Address(BaseModel):
    street_1: str | None = Field(default=None, description="The first line of the address")
    street_2: str | None = Field(default=None, description="The second line of the address")
    city: str | None = Field(default=None, description="The city of the address")
    state: str | None = Field(default=None, description="The state of the address")
    zip: str | None = Field(default=None, description="The zip code of the address")
    country: str | None = Field(default=None, description="The country of the address")
    address_type: AddressType | None = Field(default=None, description="The type of address. Either 'BILLING', 'SHIPPING', or 'OTHER'")


class Email(BaseModel):
    email: str = Field(description="The email address")
    email_type: EmailType = Field(description="The type of email. Either 'WORK', 'PERSONAL', or 'OTHER'")


class PhoneNumber(BaseModel):
    phone_number: str = Field(description="The phone number")
    phone_type: PhoneType = Field(description="The type of phone. Either 'WORK', 'PERSONAL', or 'OTHER'")


class Contact(CanonicalModel):
    first_name: str | None = Field(description="The first name of the contact")
    last_name: str | None = Field(description="The last name of the contact")
    addresses: list[Address] = Field(description="The addresses of the contact")
    emails: list[Email] = Field(description="The emails of the contact")
    phones: list[PhoneNumber] = Field(description="The phones of the contact")
    account_id: str | None = Field(description="The account id of the contact")

    @classmethod
    def stream_name(cls) -> str:
        return "crm_contacts"



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from source_faker.canon.canonical_stream import TransformFunction
from source_faker.canon.crm.contact import Address, Contact, Email, PhoneNumber

from airbyte_cdk.sources.types import Record


class UserToContact(TransformFunction[Contact]):
    def __call__(self, record: Record) -> Record:
        return Record(
            data=Contact(
                id=str(record["id"]),
                first_name=record["name"],
                last_name=None,
                addresses=[
                    Address(
                        street_number=record["address"]["street_number"],
                        street_name=record["address"]["street_name"],
                        city=record["address"]["city"],
                        state=record["address"]["state"],
                        postal_code=record["address"]["postal_code"],
                    )
                ],
                emails=[Email(email=record["email"], email_type="PERSONAL")],
                phones=[PhoneNumber(phone_number=record["telephone"], phone_type="PERSONAL")],
                account_id=None,
                created_at=record["created_at"],
                updated_at=record["updated_at"],
                deleted_at=None,
                # The address should probbaly also be in additional_properties because we're dropping the country code...
                additional_properties={
                    k: v for k, v in record.items() if k not in ["id", "name", "telephone", "created_at", "updated_at", "email"]
                },
            ).model_dump(),
            stream_name=Contact.stream_name(),
            associated_slice=record.associated_slice,
        )



#
# Copyright (c) 2023 Airbyte, Inc., all rights reserved.
#

import logging
from typing import Any, List, Mapping, MutableMapping, Optional, Tuple

from airbyte_cdk import (
    AirbyteLogFormatter,
    ConcurrentSource,
    ConcurrentSourceAdapter,
    ConfiguredAirbyteCatalog,
    ConnectorStateManager,
    Cursor,
    FinalStateCursor,
    InMemoryMessageRepository,
    Level,
    MessageRepository,
    Record,
    StreamFacade,
)
from airbyte_cdk.sources import AbstractSource
from airbyte_cdk.sources.streams import Stream
from airbyte_cdk.sources.streams.concurrent.cursor import Cursor
from airbyte_cdk.sources.streams.concurrent.partitions.partition import Partition
from source_faker.canon.canonical_stream import create_canonical_stream_facade
from source_faker.canon.crm.contact import Contact
from source_faker.canon.crm.user_to_contact import UserToContact

from .streams import Products, Purchases, Users


DEFAULT_COUNT = 1_000

logger = logging.getLogger("airbyte")


class SourceFaker(ConcurrentSourceAdapter):
    message_repository = InMemoryMessageRepository(Level(AirbyteLogFormatter.level_mapping[logger.level]))

    def __init__(self, catalog: Optional[ConfiguredAirbyteCatalog], config: Optional[Mapping[str, Any]], state: Optional[Any], **kwargs):
        concurrency_level = 1
        logger.info(f"Using concurrent cdk with concurrency level {concurrency_level}")
        concurrent_source = ConcurrentSource.create(concurrency_level, 1, logger, self._slice_logger, self.message_repository)
        super().__init__(concurrent_source)
        self.catalog = catalog
        self._state = state

    def check_connection(self, logger: logging.Logger, config: Mapping[str, Any]) -> Tuple[bool, Any]:
        if type(config["count"]) == int or type(config["count"]) == float:
            return True, None
        else:
            return False, "Count option is missing"

    def streams(self, config: Mapping[str, Any]) -> List[Stream]:
        count: int = config["count"] if "count" in config else DEFAULT_COUNT
        seed: int | None = config["seed"] if "seed" in config else None
        records_per_slice: int = config["records_per_slice"] if "records_per_slice" in config else 100
        always_updated: bool = config["always_updated"] if "always_updated" in config else True
        parallelism: int = config["parallelism"] if "parallelism" in config else 4

        state_manager = ConnectorStateManager(state=self._state)

        users_stream = self._wrap_for_concurrency(
            Users(
                self.message_repository,
                count,
                seed,
                parallelism,
                records_per_slice,
                always_updated,
                state_manager.get_stream_state("users", None),
            ),
            state_manager.get_stream_state("users", None),
            seed,
        )
        contacts_stream = create_canonical_stream_facade(
            users_stream._abstract_stream,
            Contact,
            UserToContact(),
            CanonicalFakerCursor(
                stream_name=users_stream.name,
                stream_namespace=users_stream.namespace,
                message_repository=self.message_repository,
                state=state_manager.get_stream_state("users", None),
                seed=seed,
            ),
            logger,
            self._slice_logger,
        )
        streams: List[Stream] = [
            self._wrap_for_concurrency(
                Products(
                    self.message_repository,
                    count,
                    seed,
                    parallelism,
                    records_per_slice,
                    always_updated,
                    state_manager.get_stream_state("products", None),
                ),
                state_manager.get_stream_state("products", None),
                seed,
            ),
            users_stream,
            self._wrap_for_concurrency(
                Purchases(
                    self.message_repository,
                    count,
                    seed,
                    parallelism,
                    records_per_slice,
                    always_updated,
                    state_manager.get_stream_state("purchases", None),
                ),
                state_manager.get_stream_state("purchases", None),
                seed,
            ),
            contacts_stream,
        ]

        return streams

    def _wrap_for_concurrency(self, stream: Stream, state: Any, seed: int | None):
        cursor = FakerCursor(
            stream_name=stream.name, stream_namespace=stream.namespace, message_repository=self.message_repository, state=state, seed=seed
        )
        return StreamFacade.create_from_stream(stream, self, logger, {}, cursor)


class FakerCursor(Cursor):
    def __init__(
        self, stream_name: str, stream_namespace: Optional[str], message_repository: MessageRepository, state: Any, seed: int | None
    ) -> None:
        self._stream_name = stream_name
        self._stream_namespace = stream_namespace
        self._message_repository = message_repository
        # Normally the connector state manager operates at the source-level. However, we only need it to write the sentinel
        # state message rather than manage overall source state. This is also only temporary as we move to the resumable
        # full refresh world where every stream uses a FileBasedConcurrentCursor with incremental state.
        self._connector_state_manager = ConnectorStateManager()
        self._seed = seed
        self._at_least_one_state_emitted = False
        self._state = state

    @property
    def state(self) -> MutableMapping[str, Any]:
        return self._state

    def observe(self, record: Record) -> None:
        updated_at = record.data["updated_at"]
        loop_offset = record.data["loop_offset"]
        self._state = {"seed": self._seed, "updated_at": updated_at, "loop_offset": loop_offset}
        if loop_offset % 100 == 0:
            self._emit_state_message()

    def close_partition(self, partition: Partition) -> None:
        self._emit_state_message()

    def ensure_at_least_one_state_emitted(self) -> None:
        if not self._at_least_one_state_emitted:
            self._emit_state_message()

    def _emit_state_message(self) -> None:
        self._at_least_one_state_emitted = True
        self._connector_state_manager.update_state_for_stream(self._stream_name, self._stream_namespace, self.state)
        state_message = self._connector_state_manager.create_state_message(self._stream_name, self._stream_namespace)
        self._message_repository.emit_message(state_message)


class CanonicalFakerCursor(FakerCursor):
    def __init__(
        self, stream_name: str, stream_namespace: Optional[str], message_repository: MessageRepository, state: Any, seed: int | None
    ) -> None:
        super().__init__(stream_name, stream_namespace, message_repository, state, seed)

    def observe(self, record: Record) -> None:
        updated_at = record.data["updated_at"]
        loop_offset = record.data["additional_properties"]["loop_offset"]
        self._state = {"seed": self._seed, "updated_at": updated_at, "loop_offset": loop_offset}



# Copyright (c) 2024 Airbyte, Inc., all rights reserved.

from abc import abstractmethod
from datetime import datetime
from typing import Any

from pydantic import BaseModel, Field


class TimestampedModel(BaseModel):
    created_at: datetime | None = Field(default=None, description="The date and time the record was created.")
    updated_at: datetime | None = Field(default=None, description="The date and time the record was last updated.")


class CanonicalModel(TimestampedModel):
    id: str = Field(description="The unique identifier for the record.")
    deleted_at: datetime | None = Field(default=None, description="The date and time the record was deleted.")
    additional_properties: dict[str, Any] = Field(default_factory=dict, description="Additional properties for the record.")

    @classmethod
    @abstractmethod
    def stream_name(cls) -> str:
        pass
